/*------------------------------------------------------------------------*/
/* This file provides the routine 'compsat_main' for the standard application
 * of the compsat library.  We assume that at most one file is given on the
 * command line option.  If no DIMACS file is specified stdout is copied
 * into a temporary file for further processing of the library.  This file
 * does not contain a 'main' routine, so it can be linked with the test
 * framework.
 */
#include "compsat.h"
#include "config.h"

/*------------------------------------------------------------------------*/

#include <string.h>
#include <unistd.h>

/*------------------------------------------------------------------------*/
#ifdef COMPSAT_WHITE
/*------------------------------------------------------------------------*/

int compsat_is_white_box_test_case (const char *name);
void compsat_run_white_box_test_case (CompSat * compsat, const char *name);

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static char *
generate_tmp_file_name (void)
{
  char buffer[100];

  sprintf (buffer, "/tmp/compsat-saved-stdin-%u.cnf", (unsigned) getpid ());

  return strdup (buffer);
}

/*------------------------------------------------------------------------*/

#define USAGE \
"compsat [<option> ...] [<file>]\n" \
"\n" \
"  -h                 print this command line summary\n" \
"  -p                 pretty print CNF after reading\n" \
"  --version          print version information\n" \
"  --configuration    print configuration information and exit\n" \
"  --copyright        print copyright information\n" \
"  --parse-only       parse only the input\n" \
"\n" \
"  -s|--do-not-print-assignment\n" \
"\n" \
"  <file>             input file in DIMACS format (default stdin)\n"

/*------------------------------------------------------------------------*/

static void
usage (CompSat * compsat)
{
  FILE *file;

  file = compsat_out_file (compsat);
  fprintf (file, USAGE);
  fprintf (file, "\ncore options of the COMPSAT library:\n\n");
  fprintf (file, "%s", compsat_usage ());
}

/*------------------------------------------------------------------------*/

static void
copy_file (FILE * dst, FILE * src)
{
  int ch;

  while ((ch = fgetc (src)) != EOF)
    fputc (ch, dst);
}

/*------------------------------------------------------------------------*/

int
compsat_main (CompSat * compsat, int argc, char **argv)
{
  int res, verbose, done, parse_only, pretty_print_only, print_assignment;
  CompSatArgvIterator *it;
  const char *err, *name;
  char *tmp_file_name;
  FILE *dst, *src;
  FILE *out;

#ifdef COMPSAT_WHITE
  int non_valid_tmp = 0;
  int divert_stdin = 0;
#endif

  pretty_print_only = 0;
  print_assignment = 1;
  res = COMPSAT_EXIT;
  tmp_file_name = 0;		/* use temporary file if non zero */
  parse_only = 0;
  verbose = 1;			/* SAT 2002 competition */
  name = 0;
  done = 0;			/* local options may force preliminary quit */
  err = 0;			/* error while processing options */

  /* First handle all the logging options.
   */
  it = compsat_new_ArgvIterator (compsat, argc, argv);
  while (!err && compsat_has_more_elements_ArgvIterator (it, &err))
    if (compsat_is_file_option (it->opt))
      err = compsat_option (compsat, it->opt, it->arg);

  if (err)
    compsat_report_invalid_option (compsat, it, err);

  compsat_delete_ArgvIterator (it);	/* 'err' is owned by 'it' */

  /* Then handle all other options.
   */
  it = compsat_new_ArgvIterator (compsat, argc, argv);
  while (!err && !done && compsat_has_more_elements_ArgvIterator (it, &err))
    {
      if (compsat_is_file_option (it->opt))
	continue;

      /* We first handle options specific to this executable.  These are the
       * options called 'local' above.
       */
      if (!strcmp (it->opt, "-h"))
	{
	  usage (compsat);
	  done = 1;
	}
      else if (!strcmp (it->opt, "--configuration"))
	{
	  fprintf (compsat_out_file (compsat), "%s", compsat_configuration ());
	  done = 1;
	}
      else if (!strcmp (it->opt, "--version"))
	{
	  fprintf (compsat_out_file (compsat), "%s\n", compsat_version ());
	  done = 1;
	}
      else if (!strcmp (it->opt, "--copyright"))
	{
	  fprintf (compsat_out_file (compsat), "%s\n", compsat_copyright ());
	  done = 1;
	}
      else if (!strcmp (it->opt, "-p"))
	{
	  pretty_print_only = 1;
	}
      else if (!strcmp (it->opt, "--parse-only"))
	{
	  parse_only = 1;
	}
      else if (!strcmp (it->opt, "-s") ||
	       !strcmp (it->opt, "--do-not-print-assignment"))
	{
	  print_assignment = 0;
	}
#ifdef COMPSAT_WHITE
      else if (!strcmp (it->opt, "-W"))
	{
	  if (!strcmp (it->arg, "nonvalidtmp"))
	    {
	      non_valid_tmp = 1;
	      continue;
	    }

	  if (!strcmp (it->arg, "divert"))
	    {
	      divert_stdin = 1;
	      continue;
	    }

	  if (compsat_is_white_box_test_case (it->arg))
	    {
	      compsat_run_white_box_test_case (compsat, it->arg);
	      done = 1;
	    }
	  else
	    err = "non valid white box test case";
	}
#endif
      else if (it->opt[0] != '-')
	{
	  if (name)
	    err = "can not open two files";
	  else
	    name = it->opt;
	}
      else
	{
	  /* Here we process all the common CompSat options, which are those
	   * explained in 'compsat_usage', except the logging options.
	   */
	  if (!strcmp (it->opt, "-v") || !strcmp (it->opt, "-V"))
	    verbose = 1;

	  err = compsat_option (compsat, it->opt, it->arg);
	}
    }

  if (err)
    compsat_report_invalid_option (compsat, it, err);

  compsat_delete_ArgvIterator (it);	/* 'err' is owned by 'it' */

  if (done)
    return COMPSAT_EXIT;

  if (err)
    return COMPSAT_ABORT;

  if (!name)
    {
      tmp_file_name = generate_tmp_file_name ();
      dst = fopen (tmp_file_name, "w");
#ifdef COMPSAT_WHITE
      if (dst && non_valid_tmp)	/* simulate non valid tmp file */
	{
	  fclose (dst);
	  unlink (tmp_file_name);
	  dst = 0;
	}
#endif
      if (!dst)
	{
	  fprintf (compsat_err_file (compsat),
		   "*** could not open temporary file\n");
	  free (tmp_file_name);
	  return COMPSAT_ABORT;
	}

#ifdef COMPSAT_WHITE
      if (divert_stdin)
	src = fopen ("log/divert.in", "r");
      else
#endif
	src = stdin;
      copy_file (dst, src);
      fclose (dst);

#ifdef COMPSAT_WHITE
      if (divert_stdin)
	fclose (src);
#endif
      name = tmp_file_name;
    }

  err = compsat_read (compsat, name);

  if (tmp_file_name)
    {
      unlink (tmp_file_name);
      free (tmp_file_name);
    }

  if (err)
    {
      fprintf (compsat_err_file (compsat), "*** %s\n", err);
      res = COMPSAT_ABORT;
    }
  else if (parse_only)
    {
      res = COMPSAT_EXIT;
    }
  else if (pretty_print_only)
    {
      compsat_print (compsat, compsat_out_file (compsat));
      res = COMPSAT_EXIT;
    }
  else
    {
      res = compsat_sat (compsat);
      out = compsat_out_file (compsat);

      /* SAT 2003 competition format.
       */
      if (res == COMPSAT_SATISFIABLE)
	fputs ("s SATISFIABLE\n", out);
      else if (res == COMPSAT_UNSATISFIABLE)
	fputs ("s UNSATISFIABLE\n", out);
      else
	{
	  fputs ("s UNKNOWN\n", out);
	  fprintf (out, "c %s\n", compsat_res2str (res));
	}

      if (res == COMPSAT_SATISFIABLE && print_assignment)
	compsat_print_assignment (compsat, out);
    }

  if (verbose)
    compsat_stats (compsat, compsat_log_file (compsat));

  return res;
}
